#include "plumpconfig.h"

#define _STRUCTURED_PROC 1
#include <sys/procfs.h>

typedef struct _plump_procfs_t {

	int controlFd;
	int addressSpaceFd;

} plump_procfs_t;

int procfs_okay()
{
#ifdef MEDIUM_HAS_PROCFS
	return PLUMP_SUCCESS;
#else
	return PLUMP_ERROR_NOTSUPPORTED;
#endif
}

plumpmedium_t procfs_getMedium()
{
	return PLUMP_MEDIUM_PROCFS;
}

#ifdef MEDIUM_HAS_PROCFS

int procfs_open(plump_t *p, int pid, unsigned long access)
{
	plump_procfs_t *ext = NULL;
	int ret = PLUMP_ERROR_FAILURE;
	char path[512];
	long cmd;

	memset(path, 0, sizeof(path));

	do
	{
		if (!(p->mediumExtension = malloc(sizeof(plump_procfs_t))))
			break;

		ext = (plump_procfs_t *)p->mediumExtension;

		// Open the control file
		snprintf(path, sizeof(path) - 1, "/proc/%d/ctl", pid);

		if ((ext->controlFd = open(path, O_WRONLY)) < 0)
		{
			ret = PLUMP_ERROR_NOSUCHPID;
			break;
		}

		// Open the address space file
		snprintf(path, sizeof(path) - 1, "/proc/%d/as", pid);

		if ((ext->addressSpaceFd = open(path, O_RDWR)) < 0)
		{
			ret = PLUMP_ERROR_NOSUCHPID;
			break;
		}

		cmd = PCSTOP;

		// Instruct the process to stop
		if (write(ext->controlFd, &cmd, sizeof(cmd)) < 0)
		{
			ret = PLUMP_ERROR_FAILURE;
			break;
		}

		ret = PLUMP_SUCCESS;

	} while (0);

	if (ret != PLUMP_SUCCESS)
	{
		if (ext)
		{
			if (ext->controlFd > 0)
				close(ext->controlFd);
			if (ext->addressSpaceFd > 0)
				close(ext->addressSpaceFd);

			free(ext);
		}

		p->mediumExtension = NULL;
	}

	return ret;
}

int procfs_close(plump_t *p)
{
	plump_procfs_t *ext = p->mediumExtension;
	int ret = PLUMP_SUCCESS;

	if (ext)
	{
		if (ext->controlFd > 0)
			close(ext->controlFd);
		if (ext->addressSpaceFd > 0)
			close(ext->addressSpaceFd);

		free(ext);
	}
		
	p->mediumExtension = NULL;

	return ret;
}

int procfs_execContinue(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags)
{
	long startcmd[2] = { PCRUN, PRCSIG | PRCFAULT };
	int ret = PLUMP_ERROR_FAILURE;

	do
	{
		// If a start address was specified, set the %pc register to it
		if (start)
		{
			long setvaddr[2] = { PCSVADDR, (long)start };

			if (write(p->controlFd, setvaddr, sizeof(setvaddr)) < 0)
				break;
		}

		// If we're singlestepping, set the step flag
		if (flags & PLUMP_CONTINUE_FLAG_SINGLESTEP)
			startcmd[1] |= PRSTEP;

		if (write(p->controlFd, startcmd, sizeof(startcmd)) < 0)
			break;

		ret = PLUMP_SUCCESS;
	
		// add code for detecting stop state

	} while (0);

	return ret;
}

int procfs_execStop(plump_t *p)
{
	plump_procfs_t *ext = p->mediumExtension;
	int ret;
	long cmd = PCSTOP;

	// Is it already stopped?
	if (p->state & PLUMP_STATE_STOPPED)
		return PLUMP_ERROR_ALREADY;

	// Send the PCSTOP command to the control fd
	if (write(ext->controlFd, &cmd, sizeof(long)) < 0)
	{
		switch (errno)
		{
			case EACCES:
				ret = PLUMP_ERROR_NOACCESS;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
	{
		ret       = PLUMP_SUCCESS;
		p->state |= PLUMP_STATE_STOPPED;
	}

	return ret;
}

int procfs_memoryWrite(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesWritten)
{
	plump_procfs_t *ext = p->mediumExtension;
	int ret, written;
	
	if (bytesWritten)
		*bytesWritten = 0;

	if ((written = pwrite(ext->addressSpaceFd, buffer, bufferLength, (int)addr)) < 0)
		ret = PLUMP_ERROR_FAILURE;
	else
		ret = PLUMP_SUCCESS;

	if (bytesWritten && ret == PLUMP_SUCCESS)
		*bytesWritten = written;

	return ret;
}

int procfs_memoryRead(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesRead)
{
	plump_procfs_t *ext = p->mediumExtension;
	int ret, actual;
	
	if (bytesRead)
		*bytesRead = 0;

	if ((actual = pread(ext->addressSpaceFd, buffer, bufferLength, (int)addr)) < 0)
		ret = PLUMP_ERROR_FAILURE;
	else
		ret = PLUMP_SUCCESS;

	if (bytesRead && ret == PLUMP_SUCCESS)
		*bytesRead = actual;

	return ret;
}

int procfs_memoryIsValid(plump_t *p, plumpaddr_t addr)
{
	char b;

	return procfs_memoryRead(p, addr, &b, 1, NULL);
}

int procfs_memoryGetEndOfSegment(plump_t *p, plumpaddr_t addr, plumpaddr_t *end)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	return ret;
}

int procfs_registersSet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	return ret;
}

int procfs_registersGet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	return ret;
}

#else

int procfs_open(plump_t *p, int pid, unsigned long access)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_close(plump_t *p)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_execContinue(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_execStop(plump_t *p)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_memoryWrite(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_memoryRead(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_memoryIsValid(plump_t *p, plumpaddr_t addr)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_memoryGetEndOfSegment(plump_t *p, plumpaddr_t addr, plumpaddr_t *end)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_registersSet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

int procfs_registersGet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	return PLUMP_ERROR_NOTSUPPORTED;
}

#endif
