/*
    dltrace 0.5
    Copyright (C) 2005 Richard Johnson <rjohnson@idefense.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "dltrace.h"
#include "libdasm.h"

int dt_init_trace(proc_t *ps)
{
	plumpreg_t *regs = ps->regs;

	if(options.attach < 1)
	{
		/* skip to program entry point so shard libs get loaded */
		plump_breakpointAdd(ps->plump, ps->e_entry);
		do
		{
			if(!(plump_execContinue(ps->plump, 0, NULL, PLUMP_CONTINUE_FLAG_UNTILUSRBP) == PLUMP_SUCCESS))
			{
				vlog(stderr, "fatal error: could not find program entry point\n");
				return -1;
			}
			
			if(!(plump_registersGetProgramCounter(ps->plump, &regs[8]) == PLUMP_SUCCESS))
			{
				vlog(stderr, "fatal error: could not retrieve program counter register\n");
				return -1;
			}
			
		} 
		while(EIP < ps->e_entry);
	}

	/* initialize enabled trace modes */
	if(options.ltrace)
	{
		/* locate library load addrs and relocate shared lib symbols */
		if((dt_library_reloc_syms(ps)) < 0)
		{
			vlog(stdout, "fatal error: could not relocate library symbols\n");
			return -1;
		}

		/* add breakpoints for the shared library call trace */
		if((dt_init_ltrace(ps)) < 0)
		{
			vlog(stdout, "fatal error: could not initialize library trace\n");
			return -1;
		}

	}

#if 0
	if(options.strace)
		dt_init_strace(ps);
	if(options.btrace)
		dt_init_btrace(ps);
#endif		

	return 0;
}


int dt_run_trace(proc_t *ps)
{
	int ret;
	plump_t *plump = ps->plump;
	
	while((ret = plump_execContinue(plump, 0, NULL, PLUMP_CONTINUE_FLAG_UNTILUSRBP)) == PLUMP_SUCCESS)
	{
		plump_registersGet(plump, ps->regs, 15);
		if(EIP == 0 || EIP == -1)
			break;

		/* execute analysis routines for each trace mode */
		if(options.ltrace)
			if(dt_run_ltrace(ps) < 0)
				return -1;
#if 0		
		if(options.strace)
			if(dt_run_strace(ps) < 0)
				return -1;
		if(options.btrace)
			if(dt_run_btrace(ps) < 0)
				return -1;
#endif	
	}

	vlog(stdout, "\n\n");
	return ret;
}


/* add breakpoints/create trees for the shared library call trace */
int32_t dt_init_ltrace(proc_t *ps)
{
	uint32_t addr;
	breakpoint_t *bp = NULL;
	symbol_t *symbol = NULL;
	list_t *s = NULL;
	list_t *l = NULL;
	
	ps->depth = 0;
	ps->shlib_functions = splay_init();

	/* add shared library function symbols to splay tree */
	l = ps->liblist;
	while(l->next != NULL)
	{
		s = ((library_t *)l->data)->symbols;
		while(s->next != NULL)
		{
			symbol = (symbol_t *)s->data;
			addr = symbol->addr;

			if(!(plump_breakpointAdd(ps->plump, addr) == PLUMP_SUCCESS))
				if(options.verbose > 1)
					vlog(stderr, "warning: could not add breakpoint at 0x%08x\n", addr);

			bp = malloc(sizeof(breakpoint_t));
			if(bp < 0)
			{
				perror("malloc");
				return -1;
			}
			memset(bp, 0, sizeof(breakpoint_t));
	
			bp->symbol = symbol;
			bp->addr = addr; 
			
			splay_insert(&ps->shlib_functions, (void *)addr, bp, sym_find_by_addr); 
			if(options.verbose > 2)
				vlog(stderr, "dt_init_ltrace: breakpoint added for 0x%08x: %s\n", bp->addr, bp->symbol->name);
			
			s = s->next;
		}
		l = l->next;
	}

	/* walk symbol lists and analyze code segments. this is delayed because symbols are referenced 
	in the analysis function and need to all be loaded before xrefs can be done */
	l = ps->liblist;
	while(l->next != NULL)
	{
		s = ((library_t *)l->data)->symbols;
		while(s->next != NULL)
		{
			symbol = s->data;	
			if(!symbol->analyzed)
			{
				if(dt_function_analyze(ps, symbol->addr, symbol->size) < 0)
				{
					vlog(stdout, "error: could not analyze %s()\n", symbol->name);
					return -1;
				}
				symbol->analyzed = 1;
			}

			s = s->next;
		}
		l = l->next;
	}	

	return 0;
}


/* 
This function analyzes the target process's code to determine the number of arguments.
Rather than detect prologues and check for offsets from ebp/esp, we do some basic 
heuristics from the calling function to determine arg count as they're placed on the stack
There are a few issues with this method, however its fairly accurate and will do for now. 
*/

int dt_function_analyze(proc_t *ps, uint32_t addr, uint32_t len)
{
	INSTRUCTION inst;
	breakpoint_t *bp = NULL;
	symbol_t *symbol = NULL;
	uint8_t *buf, *ptr; 
	uint32_t insn_len = 0, argn = 0, vaddr, pushn = 0;
	char string[64];

	/* copy function into buf for disassembly */
	if(!(buf = dt_mem_get_chunk(ps, addr, len)))
	{
		if(options.verbose > 1)
			vlog(stderr, "dt_function_analyze: could not retrieve memory from 0x%08x for analysis\n", addr);
		return -1;
	}

	/* disassemble to determine argc */
	for(ptr = buf; ptr <= buf + len; ptr += insn_len)
	{
		bp = NULL;
		vaddr = addr + (ptr - buf);

		insn_len = get_instruction(&inst, ptr, MODE_32);
		if(insn_len == 0)
			break;
		
		if((inst.type == INSTRUCTION_TYPE_MOV) && (inst.op1.basereg == REGISTER_ESP))
		{
				get_instruction_string(&inst, FORMAT_ATT, 0, string, sizeof(string) - 1);
				if(argn < (inst.op1.displacement / 4 + 1) && (inst.op1.displacement / 4 + 1) < 10)
					argn = inst.op1.displacement / 4 + 1;
		}
		else if(inst.type == INSTRUCTION_TYPE_PUSH)
			pushn++;
		else if (inst.type == INSTRUCTION_TYPE_CALL) 
		{
			if(inst.op1.basereg == 10)
			{
				get_instruction_string(&inst, FORMAT_ATT, 0, string, sizeof(string) - 1);
				if((bp = splay_find(&ps->shlib_functions, (void *)(vaddr + inst.op1.immediate + insn_len), sym_find_by_addr)))
				{
					if(argn == 0)
						argn = pushn;
				
					symbol = (symbol_t *)bp->symbol;
					if(symbol->argc == 0)
						symbol->argc = argn;
					else if(argn < symbol->argc && argn != 0)
					{
						if(options.verbose > 2)
							vlog(stderr, "dt_function_analyze: arg detection clash for %s() [%d args]. using: %d\n", 
								symbol->name, symbol->argc, argn);
						symbol->argc = argn;
					}
				}			
			}
			else
				pushn = argn = 0;
		}
	}
	free(buf);
	return 0;
}


// this is a bit hackish ;)
void dt_print_ltrace_ret(proc_t *ps, breakpoint_t *bp, uint32_t retaddr)
{
	uint8_t tabs[128], spacer[3] = "  ";
	symbol_t *symbol = bp->symbol;
	uint32_t i;
	
	if(options.tree)
		spacer[0] = '|';
	
	if(options.regs == 1)
	{
		memset(tabs, ' ', 50);
		tabs[50] = 0;
	}
	else if(options.regs > 1)
	{
		memset(tabs, ' ', 29);
		tabs[29] = 0;
	}

	vlog(stdout, "\n[%08x]  <|  [%08x]  ", EIP, symbol->addr);
	if(options.regs == 1)
		vlog(stdout, "%s", tabs);
	for(i = 0; i < ps->depth; i++)
		vlog(stdout, "%s", spacer);
	
	if(EAX < 65536 && EAX > -65536)
		vlog(stdout, "== %d", EAX);
	else
		vlog(stdout, "== 0x%08x", EAX);
	
	return;
}

// oh it gets better! (need to clean up output and add xml next rls)
void dt_print_ltrace_call(proc_t *ps, breakpoint_t *bp, uint32_t retaddr)
{
	uint8_t *s, tabs[64], spacer[3] = "  ";
	symbol_t *symbol = bp->symbol;
	int32_t i, arg;
	
	if(options.tree)
		spacer[0] = '|';

	memset(tabs, ' ', 29);
	tabs[28] = 0;
	
	vlog(stdout, "\n[%08x]  |>  [%08x]  ", retaddr - 5, EIP);
	if(options.regs == 1)
		vlog(stdout, "# eip=0x%08x esp=0x%08x ebp=0x%08x #  ", EIP, ESP, EBP);
	
	for(i = 0; i < ps->depth; i++)
		vlog(stdout, spacer);
	
	vlog(stdout, "%s (", symbol->name);			
	for(i = 1; i <= symbol->argc; i++)
	{
		if(!(plump_memoryRead(ps->plump, ESP + (4 * i), &arg, sizeof(uint32_t), 0) == PLUMP_SUCCESS))
			vlog(stdout, "0x%x", ESP + (4 * i));	
		else if(arg < 65536 && arg > -65536)
			vlog(stdout, "%d", arg);
		else if((s = dt_mem_get_string(ps, arg, 0)))
		{
			vlog(stdout, "0x%08x->\"%s\"", arg, s);
			free(s);
		}
		else 
		{
			vlog(stdout, "0x%08x", arg);
			if(plump_memoryRead(ps->plump, arg, &arg, sizeof(uint32_t), 0) == PLUMP_SUCCESS)
				if(arg < 65535 && arg > -65535 && arg != 0)
					vlog(stdout, "->%d", arg);
		}
		if(i < symbol->argc)
			vlog(stdout, ", ");
	}
	vlog(stdout, ")");
	
	if(options.regs > 1)
	{
		vlog(stdout, "\n%s", tabs);
		for(i = 0; i < ps->depth; i++)
			vlog(stdout, spacer);
			
		vlog(stdout, "  eax=%08x ebx=%08x ecx=%08x edx=%08x edi=%08x esi=%08x\n", 
			EAX, EBX, ECX, EDX, EDI, ESI);
			
		vlog(stdout, "%s", tabs);
		for(i = 0; i < ps->depth; i++)
			vlog(stdout, spacer);
			
		vlog(stdout, "  eip=%08x esp=%08x ebp=%08x",
			EIP, ESP, EBP);
	}			
	if(options.regs > 2)	
	{
		vlog(stdout, "\n%s", tabs);
		for(i = 0; i < ps->depth; i++)
			vlog(stdout, spacer);
		vlog(stdout, "  cs=%04x  ss=%04x  ds=%04x  es=%04x  fs=%04x  gs=%04x",
			CS, SS, DS, ES, FS, GS);
	}
}


int dt_run_ltrace(proc_t *ps) 
{
	breakpoint_t *ret_bp = NULL, *bp = NULL;
	uint32_t retaddr;
	plump_t *plump = ps->plump;

	if(!(plump_memoryRead(ps->plump, ESP, &retaddr, sizeof(uint32_t), 0) == PLUMP_SUCCESS))
	{
		vlog(stdout, "dt_run_ltrace: fatal error. could not retrieve return address\n");
		return -1;
	}

	/* shared lib call return */
	if(ps->callstack && EIP == ((breakpoint_t *)ps->callstack->data)->addr)
	{
		bp = (breakpoint_t *)list_pop(&ps->callstack);
		
		ps->depth--;
		if(ps->depth <= options.depth)
		{
			if(ps->last_call == 1 && options.regs < 2)
			{
				if(EAX < 65536)
					vlog(stdout, "\t\t==  %d", EAX);
				else
					vlog(stdout, "\t\t==  0x%08x", EAX);

			}
			else	
				dt_print_ltrace_ret(ps, bp, retaddr);
		
			ps->last_call = 0;
		}
	} 
	/* shared lib call */
	else if((bp = splay_find(&ps->shlib_functions, (void *)EIP, sym_find_by_addr)))
	{
		symbol_t *symbol = bp->symbol;
	
		if(ps->depth <= options.depth)
		{
			dt_print_ltrace_call(ps, bp, retaddr);
		}			
		
		plump_breakpointAdd(plump, retaddr);		
		ret_bp = malloc(sizeof(breakpoint_t));
		if(!ret_bp)
		{
			perror("malloc");
			return -1;
		}
		ret_bp->addr = retaddr;	
		ret_bp->symbol = symbol;	
		list_push(&ps->callstack, ret_bp);

		ps->last_call = 1;
		ps->depth++;
	} 

	return 0;
}

